source("klminimizer.R")
library(rhlra)

inv_ac_diags_routine <- inv_ac_diags

model_noise <- function(N, sigma = 1) {
    rnorm(N) * sigma
}

cadzow_mixed_alphas_factory <- function(N, r, L, alphas, iterations, arp_coefs = numeric(0), cadzow_eps = 1e-6) {
    K <- N - L + 1

    W0_diags <- inv_ac_diags_routine(N, arp_coefs)

    left_diags <- inv_ac_diags_routine(L, arp_coefs)

    right_matrices <- sapply(alphas, function(alpha) {
        list(kl_full_optimw(W0_diags, left_diags, alpha))
    }, simplify = FALSE)

    return(function(series) {
        it_actual <- numeric(length(iterations))

        for (i in seq_along(iterations)) {
            it <- iterations[i]

            cadzow_pars <- list(cadzow_epsilon = 0,
                cadzow_it_limit = it,
                svd_type = "svd")

            if (is.infinite(it)) {
                cadzow_pars <- list(cadzow_epsilon = cadzow_eps,
                    cadzow_it_limit = 500,
                    svd_type = "svd")
            }

            hlra_result <- hlra_cadzow(series, r, L, left_diags = left_diags,
                                  right_diags = right_matrices[[i]],
                                  additional_pars = cadzow_pars)
            it_actual[i] <- hlra_result$it
            series <- hlra_result$signal
        }

        list(series = series, it_actual = it_actual)
    })
}

default_signal <- sin((1:40) * 2 * pi / 6)

mixed_cadzow_example <- function(I_1 = 0:10, alpha_1 = 0.1, sample_size = 100,
                                 sigma = 0.1, cl = NULL) {
    N <- length(default_signal)
    L <- 20
    signal <- default_signal
    r <- 2

    set.seed(15)
    sample <- sapply(1:sample_size, function(x) {
        model_noise(N, sigma) + signal
    }, simplify = FALSE)

    alpha_2 <- 1
    I_2 <- 5

    sapply_ns_routine <- function(x, fun, ...) sapply(x, fun, ..., simplify = FALSE)

    if (!is.null(cl)) {
        sapply_ns_routine <- function(x, fun, ...) parSapply(cl, x, fun, ..., simplify = FALSE)
    }

    results <- sapply(I_1, function(I) {
        print(I)
        cadzow_current <- cadzow_mixed_alphas_factory(N, r, L, c(alpha_1, alpha_2), c(I, I_2))
        estimates <- sapply_ns_routine(sample, cadzow_current)
        squared_differences <- sapply(estimates, function(cadzow_result) {
            mean((cadzow_result$series - signal)^2)
        })

        iterations <- sapply(estimates, function(cadzow_result) {
            sum(cadzow_result$it_actual)
        })

        list(mse = mean(squared_differences), mse_sd = sd(squared_differences),
             avg_iterations = mean(iterations), iterations_sd = sd(iterations),
             rmse = sqrt(mean(squared_differences)))
    })

    results <- data.frame(t(results))

    plot(I_1, results$rmse, type = "b")
    plot(I_1, results$avg_iterations, type = "b")

    results
}

mixed_cadzow_example_multiple_alpha <- function(sample_size = 100, cl = NULL) {
    I_1 <- 0:15
    alphas <- c(0.5, 0.1, 0.01)
    results <- sapply(alphas, function(alpha)
        as.numeric(mixed_cadzow_example(I_1, alpha, sample_size = sample_size, cl = cl)$rmse))
    matplot(I_1, results, type = "b", pch=1:length(alphas))
    legend("topright", legend = alphas, pch=1:length(alphas),
           col = 1:length(alphas))
    results
}

make_cluster <- function(cores = 8) {
    library(snow)
    cl <- makeCluster(getOption("cl.cores", cores))
    clusterCall(cl, function() library(svd))
    clusterCall(cl, function() library(Matrix))
    clusterCall(cl, function() library(rhlra))
    clusterExport(cl, ls(envir = globalenv()))
    cl
}

slow_fast_convergence_example <- function(
                                 signal = default_signal,
                                 L = 20,
                                 r = 2,
                                 I_1 = 1,
                                 I_sum = 1:10,
                                 alpha_1 = 0.1,
                                 sample_size = 100,
                                 sigma = 0.1,
                                 cl = NULL) {
    N <- length(signal)

    set.seed(15)
    sample <- sapply(1:sample_size, function(x) {
        model_noise(N, sigma) + signal
    }, simplify = FALSE)

    alpha_2 <- 1
    I_2 <- 5

    sapply_ns_routine <- function(x, fun, ...) sapply(x, fun, ..., simplify = FALSE)

    if (!is.null(cl)) {
        sapply_ns_routine <- function(x, fun, ...) parSapply(cl, x, fun, ..., simplify = FALSE)
    }

    results <- sapply(I_sum, function(I) {
        print(I)
        I_1 <- min(I_1, I)
        I_2 <- max(0, I - I_1)
        cadzow_current <- cadzow_mixed_alphas_factory(N, r, L, c(alpha_1, alpha_2), c(I_1, I_2))
        estimates <- sapply_ns_routine(sample, cadzow_current)
        squared_differences <- sapply(estimates, function(cadzow_result) {
            mean((cadzow_result$series - signal)^2)
        })

        iterations <- sapply(estimates, function(cadzow_result) {
            sum(cadzow_result$it_actual)
        })

        list(mse = mean(squared_differences), mse_sd = sd(squared_differences),
             avg_iterations = mean(iterations), iterations_sd = sd(iterations),
             rmse = sqrt(mean(squared_differences)))
    })

    results <- data.frame(t(results))

    plot(I_sum, results$rmse, type = "b")
    plot(I_sum, results$avg_iterations, type = "b")

    results
}

default_I_1 <- c(0, 2, 4, 10, 20)
default_I_sum <- 1:20

slow_fast_convergence_multiple_I <- function(cl = NULL,
                                             signal = default_signal,
                                             L = 20,
                                             r = 2,
                                             I_1 = default_I_1,
                                             I_sum = default_I_sum,
                                             alpha = 0.02,
                                             sample_size = 100) {
    results <- sapply(I_1, function(I)
        as.numeric(slow_fast_convergence_example(signal = signal,
                                                 L = L,
                                                 r = r,
                                                 I_1 = I,
                                                 I_sum = I_sum,
                                                 alpha = alpha,
                                                 sample_size = sample_size,
                                                 cl = cl)$rmse))
    matplot(I_sum, results, type = "b", pch=1:length(I_1), lty = 1,
            xlab = "Iteration", ylab = "RMSE")
    legend("topright",
           legend = paste(alpha, "x", I_1, "+", max(I_sum)-I_1, "x", 1),
           pch=1:length(I_1),
           col = 1:length(I_1))
    results
}

# without cluster
# results <- mixed_cadzow_example_multiple_alpha()

# with cluster
# cl <- NULL
# cl <- make_cluster()
# results <- mixed_cadzow_example_multiple_alpha(cl, sample_size = 1000)
# stopCluster(cl)
